// ==============================================================
//                 ORBITER MODULE: gps
//                  
//
// gps.cpp
// Control module for gps vessel class
//
// 
// ==============================================================

#define STRICT
#define ORBITER_MODULE

#include "orbitersdk.h"
#include <cstdio>
#include <cmath>
#include <string>
#include <stdio.h>

// ==============================================================
// Some vessel parameters
// ==============================================================

const double EMPTYMASS = 2270;
const double FUELMASS = 1610;
const double PA_ISP = 2800;
const double PB_MAXRCSTH = 5;
const double PB_MAXRCSTH10 = 25;
const double PB_MAXRCSTH20 = 50;
const double ARRAY_OPERATING_SPEED = 0.05;
const double ANTENNA_OPERATING_SPEED = 0.05;
int doprograde = 0;
int lock = 0;
int a;



// Calculate lift coefficient [Cl] as a function of aoa (angle of attack) over -Pi ... Pi
// Implemented here as a piecewise linear function
double LiftCoeff (double aoa)
{
	const int nlift = 9;
	static const double AOA[nlift] = {-180*RAD,-60*RAD,-30*RAD,-1*RAD,15*RAD,20*RAD,25*RAD,60*RAD,180*RAD};
	static const double CL[nlift]  = {       0,      0,		0,		0,		0,		0,		0,     0,      0};
	static const double SCL[nlift] = {(CL[1]-CL[0])/(AOA[1]-AOA[0]), (CL[2]-CL[1])/(AOA[2]-AOA[1]),
		                              (CL[3]-CL[2])/(AOA[3]-AOA[2]), (CL[4]-CL[3])/(AOA[4]-AOA[3]),
									  (CL[5]-CL[4])/(AOA[5]-AOA[4]), (CL[6]-CL[5])/(AOA[6]-AOA[5]),
									  (CL[7]-CL[6])/(AOA[7]-AOA[6]), (CL[8]-CL[7])/(AOA[8]-AOA[7])};
	int i;
	for (i = 0; i < nlift-1 && AOA[i+1] < aoa; i++);

	return CL[i] + (aoa-AOA[i])*SCL[i];
}


class gps: public VESSEL3 {
public:
	gps (OBJHANDLE hVessel, int flightmodel);
	~gps ();

	enum DoorStatus { DOOR_CLOSED, DOOR_OPEN, DOOR_CLOSING, DOOR_OPENING } ant_status, fold_status;
	void DefineAnimations (void);
	void RevertAntenna (void);
	void RevertArray (void);
	void clbkSetClassCaps (FILEHANDLE cfg);
	void clbkPreStep(double SimT, double SimDT, double MJD);
	void clbkPostStep (double simt, double simdt, double mjd);
	bool clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp);
	void clbkPostCreation();
	void clbkSaveState (FILEHANDLE scn);
	void clbkLoadStateEx (FILEHANDLE scn, void *status);
	void DoPrograde();
	int clbkConsumeBufferedKey(DWORD key, bool down, char *kstate);
	

private:
	UINT anim_ant, anim_fold;
	double ant_proc, fold_proc;

	MGROUP_TRANSFORM *ArrayLFold1, *ArrayLFold2, *ArrayLFold3, *ArrayLFold4;
	MGROUP_TRANSFORM *ArrayRFold1, *ArrayRFold2, *ArrayRFold3, *ArrayRFold4;
	MGROUP_TRANSFORM *ArrayRFold5, *ArrayRFold6, *ArrayRFold7, *ArrayRFold8, *ArrayRFold9, *HiGainAnt1;

	THRUSTER_HANDLE th_rcs[35], th_group[4];
	PROPELLANT_HANDLE rcspr;
	ANIMATIONCOMPONENT_HANDLE Section, Section1, Section2, Section3, Section4, Section5;
	ANIMATIONCOMPONENT_HANDLE Section6, Section7, Section8, Section9, Section10;
};

gps::gps (OBJHANDLE hVessel, int flightmodel)
: VESSEL3 (hVessel, flightmodel)
{

}

gps::~gps ()
{
	delete ArrayLFold1;
	delete ArrayLFold2;
	delete ArrayLFold3;
	delete ArrayLFold4;
	delete ArrayRFold1;
	delete ArrayRFold2;
	delete ArrayRFold3;
	delete ArrayRFold4;
	delete ArrayRFold5;
	delete ArrayRFold6;
	delete ArrayRFold7;
	delete ArrayRFold8;
	delete ArrayRFold9;
	delete HiGainAnt1;
	
}
// Define animation sequences for moving parts
// --------------------------------------------------------------
void gps::DefineAnimations (void)
{
	// 1. HG ant
	static UINT HiGainAnt1Grp[1] = {1};
	HiGainAnt1 = new MGROUP_ROTATE(0, HiGainAnt1Grp, 0, _V(0,0,0), _V(1,0,0), (float)(PI*2));
	anim_ant = CreateAnimation (0);
	Section = AddAnimationComponent (anim_ant, 0, 1, HiGainAnt1);


	
	// 2. Solar arrays - folding
	anim_fold = CreateAnimation (0);
	static UINT ArrayLFoldGrp1[1] = {17};
	static UINT ArrayLFoldGrp2[1] = {18};
	static UINT ArrayLFoldGrp3[1] = {19};
	static UINT ArrayLFoldGrp4[1] = {20};
	static UINT ArrayRFoldGrp1[1] = {24};
	static UINT ArrayRFoldGrp2[1] = {23};
	static UINT ArrayRFoldGrp3[1] = {22};
	static UINT ArrayRFoldGrp4[1] = {21};
	static UINT ArrayRAntGrp1[1] = {4};
	static UINT ArrayLAntGrp2[1] = {5};
	static UINT ArrayRAntGrp3[1] = {6};
	static UINT ArrayLAntGrp4[1] = {7};
	static UINT ArrayLAntGrp5[1] = {42};

	ArrayLFold1 = new MGROUP_ROTATE(0, ArrayLFoldGrp1, 1, _V(1.44,0,0), _V(0,1,0), (float)(PI*0.5));
	Section1=AddAnimationComponent (anim_fold, 0,   1, ArrayLFold1, Section);
	ArrayLFold2 = new MGROUP_ROTATE(0, ArrayLFoldGrp2, 1, _V(1.46,0,1.228), _V(0,-1,0), (float)(PI));
	Section2=AddAnimationComponent (anim_fold, 0,   1, ArrayLFold2, Section1);
	ArrayLFold3 = new MGROUP_ROTATE(0, ArrayLFoldGrp3, 1, _V(1.483,0,-1.228), _V(0,1,0), (float)(PI));
	Section3=AddAnimationComponent (anim_fold, 0,   1, ArrayLFold3, Section2);
	ArrayLFold4 = new MGROUP_ROTATE(0, ArrayLFoldGrp4, 1, _V(1.506,0,1.228), _V(0,-1,0), (float)(PI));
	Section4=AddAnimationComponent (anim_fold, 0,   1, ArrayLFold4, Section3);

	ArrayRFold1 = new MGROUP_ROTATE(0, ArrayRFoldGrp1, 1, _V(-1.44,0,0), _V(0,-1,0), (float)(PI*0.5));
	Section5=AddAnimationComponent (anim_fold, 0,   1, ArrayRFold1, Section);
	ArrayRFold2 = new MGROUP_ROTATE(0, ArrayRFoldGrp2, 1, _V(-1.46,0,1.228), _V(0,1,0), (float)(PI));
	Section6=AddAnimationComponent (anim_fold, 0,   1, ArrayRFold2, Section5);
	ArrayRFold3 = new MGROUP_ROTATE(0, ArrayRFoldGrp3, 1, _V(-1.483,0,-1.228), _V(0,-1,0), (float)(PI));
	Section7=AddAnimationComponent (anim_fold, 0,   1, ArrayRFold3, Section6);
	ArrayRFold4 = new MGROUP_ROTATE(0, ArrayRFoldGrp4, 1, _V(-1.505,0,1.228), _V(0,1,0), (float)(PI));
	Section8=AddAnimationComponent (anim_fold, 0,   1, ArrayRFold4, Section7);
	
	ArrayRFold5 = new MGROUP_ROTATE(0, ArrayRAntGrp1, 1, _V(0.264,-1.128,0), _V(0,0,1), (float)(PI/3));
	Section9=AddAnimationComponent (anim_fold, 0,   0.15, ArrayRFold5);
	ArrayRFold6 = new MGROUP_ROTATE(0, ArrayLAntGrp2, 1, _V(-1.42,-1.132,0), _V(0,0,1), (float)(PI/2));
	AddAnimationComponent (anim_fold, 0,   0.15, ArrayRFold6, Section9);

	ArrayRFold7 = new MGROUP_ROTATE(0, ArrayRAntGrp3, 1, _V(-0.292,1.13,0), _V(0,0,1), (float)(PI/3));
	Section10=AddAnimationComponent (anim_fold, 0,   0.15, ArrayRFold7);
	ArrayRFold8 = new MGROUP_ROTATE(0, ArrayLAntGrp4, 1, _V(1.397,1.132,0), _V(0,0,1), (float)(PI/2));
	AddAnimationComponent (anim_fold, 0,   0.15, ArrayRFold8, Section10);

	ArrayRFold9 = new MGROUP_ROTATE(0, ArrayLAntGrp5, 1, _V(0.597,-1.115,0), _V(0,0,1), (float)(120*RAD));
	AddAnimationComponent (anim_fold, 0,   0.15, ArrayRFold9);

}

void gps::RevertAntenna (void)
{
	ant_status = ((ant_status == DOOR_CLOSED || ant_status == DOOR_CLOSING) ?
		DOOR_OPENING : DOOR_CLOSING);
}

void gps::RevertArray (void)
{
	fold_status = ((fold_status == DOOR_CLOSED || fold_status == DOOR_CLOSING) ?
		DOOR_OPENING : DOOR_CLOSING);
}

// ==============================================================
// gps class interface
// ==============================================================

// ==============================================================
// Overloaded callback functions
// ==============================================================

// --------------------------------------------------------------
// Set the capabilities of the vessel class
// --------------------------------------------------------------
void gps::clbkSetClassCaps (FILEHANDLE cfg)
{
	doprograde = 0;
	lock = 0;
	ant_proc = 0.0;
	ant_status = DOOR_CLOSED;
	fold_proc = 0.0;
	fold_status = DOOR_CLOSED;
	DefineAnimations ();
	// physical specs
	SetCW (1, 1, 1, 1);
	SetWingAspect (0.7);
	SetWingEffectiveness (2.5);
	SetRotDrag (_V(1,1,1));
	if (GetFlightModel() >= 1) {
		SetPitchMomentScale (5e-4);
		SetBankMomentScale (5e-4);
	}
	SetTrimScale (0.05);
	SetPMI (_V(1,2,2));
	SetCrossSections (_V(5,10,10));
	SetSize (4.0);
	SetEmptyMass (EMPTYMASS);
	SetMeshVisibilityMode ((AddMesh ("Falcon9R_V12/gps")), MESHVIS_EXTERNAL);
	SetCameraOffset (_V(0,2,-10));
	SetLiftCoeffFunc (LiftCoeff);
	SetTouchdownPoints (_V(-1,0,-0.5), _V(1,1,-0.5), _V(1,-1,-0.5));
	SetDockParams (_V(0,0,0), _V(0,1,0), _V(0,0,1));



	// propellant resources
	rcspr = CreatePropellantResource (FUELMASS, FUELMASS);
	
	// ***************** thruster definitions *******************

	th_rcs[ 0] = CreateThruster (_V( 1,0, 1), _V(0, 1,0), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[ 1] = CreateThruster (_V( 1,0, 1), _V(0,-1,0), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[ 2] = CreateThruster (_V(-1,0, 1), _V(0, 1,0), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[ 3] = CreateThruster (_V(-1,0, 1), _V(0,-1,0), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[ 4] = CreateThruster (_V( 1,0,-1), _V(0, 1,0), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[ 5] = CreateThruster (_V( 1,0,-1), _V(0,-1,0), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[ 6] = CreateThruster (_V(-1,0,-1), _V(0, 1,0), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[ 7] = CreateThruster (_V(-1,0,-1), _V(0,-1,0), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[ 8] = CreateThruster (_V( 1,0, 1), _V(-1,0,0), PB_MAXRCSTH10, rcspr, PA_ISP);
	th_rcs[ 9] = CreateThruster (_V(-1,0, 1), _V( 1,0,0), PB_MAXRCSTH10, rcspr, PA_ISP);
	th_rcs[10] = CreateThruster (_V( 1,0,-1), _V(-1,0,0), PB_MAXRCSTH10, rcspr, PA_ISP);
	th_rcs[11] = CreateThruster (_V(-1,0,-1), _V( 1,0,0), PB_MAXRCSTH10, rcspr, PA_ISP);
	th_rcs[12] = CreateThruster (_V( 0,0,-1), _V(0,0, 1), PB_MAXRCSTH10, rcspr, PA_ISP);
	th_rcs[13] = CreateThruster (_V( 0,0, 1), _V(0,0,-1), PB_MAXRCSTH10, rcspr, PA_ISP);
	th_rcs[14] = CreateThruster (_V( 1,0, 1), _V(0, 1,0), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[15] = CreateThruster (_V(-1,0,-1), _V(0,-1,0), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[16] = CreateThruster (_V( 1,0, 1), _V(0,-1,0), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[17] = CreateThruster (_V( 1,0,-1), _V(0,-1,0), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[18] = CreateThruster (_V(-1,0, 1), _V(0, 1,0), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[19] = CreateThruster (_V(-1,0,-1), _V(0, 1,0), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[20] = CreateThruster (_V( 1,0,-1), _V(0, 1,0), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[21] = CreateThruster (_V(-1,0, 1), _V(0,-1,0), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[22] = CreateThruster (_V( 1,0, 1), _V(0, 1,0), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[23] = CreateThruster (_V( 1,0,-1), _V(0, 1,0), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[24] = CreateThruster (_V(-1,0, 1), _V(0, 1,0), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[25] = CreateThruster (_V(-1,0,-1), _V(0, 1,0), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[26] = CreateThruster (_V( 1,0, 1), _V(0,-1,0), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[27] = CreateThruster (_V( 1,0,-1), _V(0,-1,0), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[28] = CreateThruster (_V(-1,0, 1), _V(0,-1,0), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[29] = CreateThruster (_V(-1,0,-1), _V(0,-1,0), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[30] = CreateThruster (_V( 1,0, 1), _V(-1,0,0), PB_MAXRCSTH10, rcspr, PA_ISP);
	th_rcs[31] = CreateThruster (_V( 1,0,-1), _V(-1,0,0), PB_MAXRCSTH10, rcspr, PA_ISP);
	th_rcs[32] = CreateThruster (_V(-1,0, 1), _V( 1,0,0), PB_MAXRCSTH10, rcspr, PA_ISP);
	th_rcs[33] = CreateThruster (_V(-1,0,-1), _V( 1,0,0), PB_MAXRCSTH10, rcspr, PA_ISP);
	th_rcs[34] = CreateThruster (_V( 0,0,-2), _V(0,0, 1), PB_MAXRCSTH20, rcspr, PA_ISP);

	th_group[0] = th_rcs[0];
	th_group[1] = th_rcs[2];
	th_group[2] = th_rcs[5];
	th_group[3] = th_rcs[7];
	CreateThrusterGroup (th_group, 4, THGROUP_ATT_PITCHUP);

	th_group[0] = th_rcs[1];
	th_group[1] = th_rcs[3];
	th_group[2] = th_rcs[4];
	th_group[3] = th_rcs[6];
	CreateThrusterGroup (th_group, 4, THGROUP_ATT_PITCHDOWN);

	th_group[0] = th_rcs[14];
	th_group[1] = th_rcs[20];
	th_group[2] = th_rcs[21];
	th_group[3] = th_rcs[15];
	CreateThrusterGroup (th_group, 4, THGROUP_ATT_BANKLEFT);

	th_group[0] = th_rcs[16];
	th_group[1] = th_rcs[17];
	th_group[2] = th_rcs[18];
	th_group[3] = th_rcs[19];
	CreateThrusterGroup (th_group, 4, THGROUP_ATT_BANKRIGHT);

	th_group[0] = th_rcs[22];
	th_group[1] = th_rcs[23];
	th_group[2] = th_rcs[24];
	th_group[3] = th_rcs[25];
	CreateThrusterGroup (th_group, 4, THGROUP_ATT_UP);

	th_group[0] = th_rcs[26];
	th_group[1] = th_rcs[27];
	th_group[2] = th_rcs[28];
	th_group[3] = th_rcs[29];
	CreateThrusterGroup (th_group, 4, THGROUP_ATT_DOWN);

	th_group[0] = th_rcs[8];
	th_group[1] = th_rcs[11];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_YAWLEFT);

	th_group[0] = th_rcs[9];
	th_group[1] = th_rcs[10];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_YAWRIGHT);

	th_group[0] = th_rcs[30];
	th_group[1] = th_rcs[31];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_LEFT);

	th_group[0] = th_rcs[32];
	th_group[1] = th_rcs[33];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_RIGHT);

	CreateThrusterGroup (th_rcs+12, 1, THGROUP_ATT_FORWARD);
	CreateThrusterGroup (th_rcs+13, 1, THGROUP_ATT_BACK);

	CreateThrusterGroup (&th_rcs[34], 1, THGROUP_MAIN);

	SURFHANDLE tex_main = oapiRegisterExhaustTexture ("Exhaust2");
	AddExhaust (th_rcs[34], 2, 0.2, -0.25, tex_main);

	SURFHANDLE exhaust = oapiRegisterExhaustTexture ("exhaust_atrcs");
	AddExhaust(th_rcs[0], 2, 0.1, _V(0.024,1.358,0.378), _V(0,0.259,0.966),exhaust);
	AddExhaust(th_rcs[0], 2, 0.1, _V(0.024,-1.358,-0.378), _V(0,-0.259,-0.966),exhaust);

	AddExhaust(th_rcs[1], 2, 0.1, _V(0.024,-1.358,0.378), _V(0,-0.259,0.966),exhaust);
	AddExhaust(th_rcs[1], 2, 0.1, _V(0.024,1.358,-0.378), _V(0, 0.259,-0.966),exhaust);

	AddExhaust(th_rcs[13], 2, 0.1, _V(0.024,-1.358,0.378), _V(0,-0.259,0.966),exhaust);
	AddExhaust(th_rcs[13], 2, 0.1, _V(0.024,1.358,0.378), _V(0, 0.259,0.966),exhaust);
	
	AddExhaust(th_rcs[12], 2, 0.1, _V(0.024,-1.358,-0.378), _V(0,-0.259,-0.966),exhaust);
	AddExhaust(th_rcs[12], 2, 0.1, _V(0.024,1.358,-0.378), _V(0, 0.259,-0.966),exhaust);

	AddExhaust(th_rcs[22], 1, 0.05, _V(0.024,-1.457,0.157), _V(0,-1,0),exhaust);
	AddExhaust(th_rcs[22], 1, 0.05, _V(0.024,-1.457,-0.157), _V(0,-1,0),exhaust);

	AddExhaust(th_rcs[26], 1, 0.05, _V(0.024,1.457,0.157), _V(0,1,0),exhaust);
	AddExhaust(th_rcs[26], 1, 0.05, _V(0.024,1.457,-0.157), _V(0,1,0),exhaust);

	AddExhaust(th_rcs[30], 1, 0.05, _V(0.197,-1.358,0.157), _V(0.966,-0.259,0),exhaust);
	AddExhaust(th_rcs[30], 1, 0.05, _V(0.197,-1.358,-0.157), _V(0.966,-0.259,0),exhaust);
	AddExhaust(th_rcs[30], 1, 0.05, _V(0.197,1.358,0.157), _V(0.966,0.259,0),exhaust);
	AddExhaust(th_rcs[30], 1, 0.05, _V(0.197,1.358,-0.157), _V(0.966,0.259,0),exhaust);

	AddExhaust(th_rcs[32], 1, 0.05, _V(-0.149,-1.358,0.157), _V(-0.966,-0.259,0),exhaust);
	AddExhaust(th_rcs[32], 1, 0.05, _V(-0.149,-1.358,-0.157), _V(-0.966,-0.259,0),exhaust);
	AddExhaust(th_rcs[32], 1, 0.05, _V(-0.149,1.358,0.157), _V(-0.966,0.259,0),exhaust);
	AddExhaust(th_rcs[32], 1, 0.05, _V(-0.149,1.358,-0.157), _V(-0.966,0.259,0),exhaust);

	AddExhaust(th_rcs[14], 1, 0.05, _V(-0.149,-1.358,0.157), _V(-0.966,-0.259,0),exhaust);
	AddExhaust(th_rcs[14], 1, 0.05, _V(-0.149,-1.358,-0.157), _V(-0.966,-0.259,0),exhaust);
	AddExhaust(th_rcs[14], 1, 0.05, _V(0.197,1.358,0.157), _V(0.966,0.259,0),exhaust);
	AddExhaust(th_rcs[14], 1, 0.05, _V(0.197,1.358,-0.157), _V(0.966,0.259,0),exhaust);

	AddExhaust(th_rcs[16], 1, 0.05, _V(0.197,-1.358,0.157), _V(0.966,-0.259,0),exhaust);
	AddExhaust(th_rcs[16], 1, 0.05, _V(0.197,-1.358,-0.157), _V(0.966,-0.259,0),exhaust);
	AddExhaust(th_rcs[16], 1, 0.05, _V(-0.149,1.358,0.157), _V(-0.966,0.259,0),exhaust);
	AddExhaust(th_rcs[16], 1, 0.05, _V(-0.149,1.358,-0.157), _V(-0.966,0.259,0),exhaust);

	AddExhaust(th_rcs[8], 1, 0.05, _V(0.197,-1.358,0.157), _V(0.966,-0.259,0),exhaust);
	AddExhaust(th_rcs[8], 1, 0.05, _V(0.197,1.358,0.157), _V(0.966,0.259,0),exhaust);
	AddExhaust(th_rcs[8], 1, 0.05, _V(-0.149,-1.358,-0.157), _V(-0.966,-0.259,0),exhaust);
	AddExhaust(th_rcs[8], 1, 0.05, _V(-0.149,1.358,-0.157), _V(-0.966,0.259,0),exhaust);

	AddExhaust(th_rcs[9], 1, 0.05, _V(-0.149,-1.358,0.157), _V(-0.966,-0.259,0),exhaust);
	AddExhaust(th_rcs[9], 1, 0.05, _V(-0.149,1.358,0.157), _V(-0.966,0.259,0),exhaust);
	AddExhaust(th_rcs[9], 1, 0.05, _V(0.197,-1.358,-0.157), _V(0.966,-0.259,0),exhaust);
	AddExhaust(th_rcs[9], 1, 0.05, _V(0.197,1.358,-0.157), _V(0.966,0.259,0),exhaust);

}

// ==============================================================
// API callback interface
// ==============================================================

// --------------------------------------------------------------
// Vessel initialisation
// --------------------------------------------------------------
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new gps (hvessel, flightmodel);
}


//gps::~gps ()
//{
//		doprograde = 0;
//		lock = 0;
//}

// --------------------------------------------------------------
// Vessel cleanup
// --------------------------------------------------------------
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (gps*)vessel;
}
// ==============================================================
// Overloaded clbkPostCreation callback functions
// ==============================================================

void gps::clbkPostCreation ()
{
	
}


void gps::clbkSaveState (FILEHANDLE scn)
{
	char cbuf[256];

	// default vessel parameters
	VESSEL3::clbkSaveState (scn);

	// custom parameters
	oapiWriteScenario_int (scn, "SUNLOCK", lock);
	oapiWriteScenario_int (scn, "AUTOATT", doprograde);
	sprintf (cbuf, "%d %0.4f", ant_status, ant_proc);
	oapiWriteScenario_string (scn, "SPR", cbuf);
	sprintf (cbuf, "%d %0.4f", fold_status, fold_proc);
	oapiWriteScenario_string (scn, "SOLP", cbuf);
}
void gps::clbkLoadStateEx (FILEHANDLE scn, void *vs)
{
    char *line;
	while (oapiReadScenario_nextline (scn, line)) {
		if (!strnicmp (line, "SUNLOCK", 7)) {
            sscanf (line+7, "%d", &lock);
		}else if (!strnicmp (line, "AUTOATT", 7)) {
			sscanf (line+7, "%d", &doprograde);
		}else if (!strnicmp (line, "SPR", 3)) {
			sscanf (line+3, "%d%lf", &ant_status, &ant_proc);
		} else if (!strnicmp (line, "SOLP", 4)) {
			sscanf (line+5, "%d%lf", &fold_status, &fold_proc);
		}else {
            ParseScenarioLineEx (line, vs);
			// unrecognised option - pass to Orbiter's generic parser
		}
	}
	SetAnimation (anim_ant, ant_proc);
	SetAnimation (anim_fold, fold_proc);
}
int gps::clbkConsumeBufferedKey(DWORD key, bool down, char *kstate) {
  if (!down) return 0; // only process keydown events

  if (KEYMOD_SHIFT (kstate)) {

  } else { // unmodified keys
    switch (key) {
	case OAPI_KEY_V: // prograde earth
		if  (doprograde < 1) doprograde = 1;
		else if (doprograde > 0) doprograde = 0;
		return 1;
	case OAPI_KEY_G: // lock solp
		//if (lock < 1) lock = 1;
		if (fold_status == DOOR_CLOSING) RevertArray();
		else if (fold_status == DOOR_CLOSED) RevertArray();
		return 1;
	}
  }
  return 0;
}

void gps::DoPrograde()
{
	DeactivateNavmode (1);
	DeactivateNavmode (2);
	DeactivateNavmode (3);
	DeactivateNavmode (4);
	DeactivateNavmode (5);
	DeactivateNavmode (6);
	DeactivateNavmode (7);
	
	// get the earth
	char cbuf[256]="Earth";
	OBJHANDLE hRef = oapiGetGbodyByName (cbuf);
	VECTOR3 velvect;
	VECTOR3 earthpos;
	GetRelativeVel (hRef, velvect);
	GetRelativePos (hRef, earthpos);

	normalise(velvect);
	normalise(earthpos);
	
	VECTOR3 xright;
	VECTOR3 yup;
	VECTOR3 zforward;
	
	xright = crossp (earthpos,velvect);
	normalise(xright);
	zforward = -earthpos;
	normalise(zforward);
	yup = crossp (xright,zforward);
	normalise(yup);
	
	double gamma = atan2( xright.x, yup.x);
	double beta = -1*(asin(zforward.x));
	double alpha = atan2(zforward.y, zforward.z);
	SetGlobalOrientation (_V(alpha, beta, (PI/2)-gamma));
	SetAngularVel (_V(0,0,0));
}



bool gps::clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp)
{

	// draw the default HUD
	VESSEL3::clbkDrawHUD (mode, hps, skp);

	int s = hps->H;
	double d = (s*0.00130208);
	int sw = ((hps->W));
	int lw =  (int)(16*sw/1024);
	int lwoffset = sw - (18*lw);
	int hlw = (int) (lw/2);


	int roxl = 0;
	int royl = 0;

	double ds = s;
	double dsw = sw;
	double sc_ratio = ds/dsw;


	if (sc_ratio < 0.7284)
	{
	roxl = (lw*10);
	royl = (int) (-88*d);
	}



	int wd = (int)(152*d);
	int wc = (int)(168*d);
	int w0 = (int)(184*d);
	int w1 = (int) (200*d);

	//char cbuf[256];
	
	{
	skp->SetTextColor(0x0066FF66);
	if (doprograde > 0)skp->SetTextColor(0x0066FFFF);
	char cbuf[256];
	sprintf(cbuf,"V = Auto Nadir Pointing");
	skp->Text ((10+roxl), (wc+royl), cbuf,strlen(cbuf));
	}
	

	if (fold_proc == 0)
	{
	skp->SetTextColor(0x0066FF66);
	char cbuf[256];
	sprintf(cbuf,"G = Deploy Solar Panels/Antennae");
	skp->Text ((10+roxl), (w0+royl), cbuf,strlen(cbuf));
	}

	return true;
}


void gps::clbkPreStep(double SimT, double SimDT, double MJD) {

	

	if (doprograde > 0) DoPrograde();
	

/////// Sun Tracking Starts Here!!!!! ////////////////
///// This is for solar panels rotating around X-axis /////

	if (fold_proc == 1) // if panels are fully open
	{
	double Solpangle;
	VECTOR3 rsun;
	VECTOR3 rsuninv;
	VECTOR3 rsunloc;
	GetGlobalPos(rsun);	// get vessel global pos (not same as pos relative to sun, but good enough)
	rsuninv = _V(-rsun.x, -rsun.y, -rsun.z); // reverse to get sun pos relative to ship
	Global2Local(rsuninv,rsunloc); // convert to vessel frame
	Solpangle = (PI*((ant_proc*2))); // current rotation angle of panels
	double sunangle = atan2(-rsunloc.y,rsunloc.z); // current angle of sun on ZY plane relative to z axis
	if (sunangle < 0) sunangle = sunangle + (2*PI); // sanity check - angle always positive

	double angle_diff = sunangle-Solpangle; // angular difference between sun and panels

	if (angle_diff < PI && angle_diff > 0)a= 1;			// check which way the panels should rotate
	else if (angle_diff > PI && angle_diff > 0)a=-1;	// check which way the panels should rotate

	else if (angle_diff > -PI && angle_diff < 0)a=-1;	// check which way the panels should rotate
	else if (angle_diff < -PI && angle_diff < 0)a= 1;	// check which way the panels should rotate

	double max_rot = SimDT * ANTENNA_OPERATING_SPEED * 2*PI;	// check if angular difference
	if (angle_diff > 0 && angle_diff < max_rot) a = 0;			// is less than panel rotation
	else if (angle_diff < 0 && -angle_diff < max_rot) a = 0;	// during this frame

	// if a = 1 rotate one way
	// if a  = -1 rotate opposite way
	// if a  = 0 set panels exactly perpendicular to sun

	
		double db = SimDT * ANTENNA_OPERATING_SPEED;
		if (ant_proc > 1) ant_proc = (ant_proc - 1); // range always 0-1
		else if (ant_proc < 0) ant_proc = (ant_proc + 1); // range always 0-1
		else if (a == 1) ant_proc = (ant_proc+db);
		else if (a == -1) ant_proc = (ant_proc-db);
		else if (a == 0) ant_proc = sunangle/(2*PI); // set exactly perpendicular to sun
		SetAnimation (anim_ant, ant_proc); // set the animation state and we're done!
	}

/////////////////////////// Sun Tracking Code ends! /////////////////

	// Animate solar arrays
	if (fold_status >= DOOR_CLOSING) {
		double da = SimDT * ARRAY_OPERATING_SPEED;
		if (fold_status == DOOR_CLOSING) {
			if (fold_proc > 0.0) fold_proc = max (0.0, fold_proc-da);
			else                 fold_status = DOOR_CLOSED;
		} else {
			if (fold_proc < 1.0) fold_proc = min (1.0, fold_proc+da);
			else                 fold_status = DOOR_OPEN;
		}
		SetAnimation (anim_fold, fold_proc);
	}

}

void gps::clbkPostStep (double simt, double simdt, double mjd)
{


}